import sys, falcon, logging, ssl, json, os, requests, re, memcache
from wsgiref import simple_server
from wsgiref.simple_server import WSGIRequestHandler

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

ERROR_CODE = 127

config = {}

memcache_client = None

#Logging configuration
logger = logging.getLogger(__name__)
logFormatter = logging.Formatter('%(asctime)s [%(levelname)-5.5s]  %(message)s')
consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logFormatter)
logging.getLogger().addHandler(consoleHandler)

logger.setLevel(logging.DEBUG)

def get_vcpe_host_name_api(ip, auth_token):
    global config, logger

    query = "json.tve.config.networking.interfaces.ipv4-address"
    # Search by hostname
    if re.match(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$", ip) is None:
        query = "json.tve.config.hostname"
    else:
        if memcache_client.get(ip):
            return memcache_client.get(ip)

    payload = {query: ip}
    inventory_api_url = config['inventory_api_url']
    url = inventory_api_url + 'v2/compute_nodes/hostname'
    headers = {"Content-Type": 'application/json', }
    headers.update({"X-Auth-Token": auth_token})

    r = requests.get(url, headers=headers, params = payload, verify=False)
    if r.status_code == 200:
        hostname = r.text
        return hostname
    return None

def get_keystone_token(username, password):
    global config, logger
    keystone_token_request = '{"user":"%s","password":"%s"}' % (username,password)

    try:
        resp = requests.get(config['keystone_url'], data=keystone_token_request,
            headers={'Content-type': 'application/json',}, verify=False, auth=('admin', password),timeout=10)
        logger.debug("Successfully taken auth token!")
        return resp.text
    except:
        logger.error("Unable to get auth token!")

    return None

def get_keystone_token_credentials(token):
    global config, logger
    headers = { 'X-Auth-Token': token, 'X-Subject-Token': token }
    data = requests.get(config['keystone_url'], headers=headers, verify=False)
    if falcon.get_http_status(data.status_code) != falcon.HTTP_200:
        logger.error(data.text.encode('utf-8', 'replace'))
        return None
    return data.text.encode('utf-8', 'replace')

class Authenticate(object):
    def on_get(self, req, resp):
        raise falcon.HTTPError(falcon.HTTP_METHOD_NOT_ALLOWED, 'GET Method is not supported for authenticate!')

    def on_post(self, req, resp):
        if not 'authentication' in config:
            raise falcon.HTTPError(falcon.HTTP_UNAUTHORIZED, 'Authentication is not enabled!')

        try:
            data = json.loads(req.stream.read(req.content_length or 0))
        except:
            raise falcon.HTTPBadRequest('Error parsing the request!')

        if 'token' in data:
            data = get_keystone_token_credentials(data['token'])
            if data is None:
                raise falcon.HTTPError(falcon.HTTP_UNAUTHORIZED, 'Error getting credentials for token "{}"!'.format(data['token']))

            resp.status = falcon.HTTP_200
            resp.content_type = 'application/json'
            resp.body = data
            return

        if not 'username' in data or not 'password' in data:
            raise falcon.HTTPBadRequest('Incorrect parameters passed!')

        if config['authentication'] == 'keystone':
            keystone_token = get_keystone_token(data['username'], data['password'])
            if keystone_token is None:
                raise falcon.HTTPError(falcon.HTTP_UNAUTHORIZED, 'Error getting authentication token for user "{}"!'.format(data['username']))

            resp.status = falcon.HTTP_200
            resp.content_type = 'application/text'
            resp.body = keystone_token
            return

        raise falcon.HTTPBadRequest('Unsupported authentication method!')

class SinkAdapter(object):
    def __call__(self, req, resp):
        global config, logger

        tokens = str(req.path).split('/')

        resp._headers = {
            'Access-Control-Allow-Origin': '*',
            'Access-Control-Allow-Methods': 'GET, PUT, POST, DELETE, PATCH, OPTIONS',
            'Access-Control-Allow-Headers': 'X-Auth-Token, Content-Type'
        }

        if len(tokens) < 2:
            raise falcon.HTTPBadRequest('No route specified!')

        #Note that the request stream can be read only once
        req_body = req.stream.read(req.content_length or 0)

        params = {}
        for wrapper in config['wrapper']:
            if tokens[1] in wrapper.keys():
                #params = wrapper.values()[0]
                #In Python3 object is not subscriptable like a list
                params = list(wrapper.values())[0]
                break

        if len(params) == 0:
            raise falcon.HTTPBadRequest('Route {} not found!'.format(str(req.path)))

        if not 'enabled' in params or params['enabled'] != 'true':
            raise falcon.HTTPBadRequest('Route {} is not enabled!'.format(str(req.path)))

        if 'operations' in params:
            methods = str(params['operations']).replace(' ', '').split(',')
            if not str(req.method) in methods:
                raise falcon.HTTPBadRequest('{} method is not allowed!'.format(str(req.method)))

        if 'regexp_url' in params:
            if re.match(params['regexp_url'], str(req.path), re.M | re.I) is None:
                raise falcon.HTTPBadRequest('The request url {} is not allowed!'.format(str(req.path)))

        if 'regexp_body' in params:
            if re.match(params['regexp_body'], str(req_body), re.M | re.I) is None:
                raise falcon.HTTPBadRequest('The request body is not allowed!')

        if req.method in ('POST', 'PUT', 'PATCH') and req.content_length in (0, None):
            raise falcon.HTTPBadRequest('Empty request body', 'A valid JSON document is required.')

        #Add the url params
        cut_pos = str(req.url).find(str(config['port']) + '/' + tokens[1]) + len(str(config['port']) + '/' + tokens[1])

        if cut_pos < 0:
            raise falcon.HTTPBadRequest('Not properly resolved url {} for configured {}'.format(req.url,config['port']))

        url = params['url'] + str(req.url)[cut_pos:]

        logger.debug('{} {} via {}'.format(req.method,url,req.url))
        logger.debug('HEADERS: ' + str(req.headers))
        logger.debug('BODY: ' + str(req_body))

        headers = req.headers
        if 'blocked_headers' in params:
            blocked_headers = str(params['blocked_headers']).replace(' ', '').split(',')
            for h_name,h_value in req.headers.items():
                if h_name in blocked_headers:
                    logger.debug('Removed {}:{} from request header'.format(h_name,h_value))
                    headers.pop(h_name)

        if str(req.method) != 'OPTIONS' and 'authentication' in config:
            if 'skip_authentication' in params and params['skip_authentication'] == 'true':
                logger.debug('Authentication skipped for {}'.format(str(req.path)))
            else:
                if config['authentication'] == 'keystone':
                    keystone_token = req.get_header('X-Auth-Token')
                    if keystone_token is None:
                        if 'vCPE_validate' in params and params['vCPE_validate'] == 'true':
                            if get_vcpe_host_name_api(str(req.remote_addr), keystone_token) is None:
                                logger.debug('RESP {} {}'.format(falcon.HTTP_UNAUTHORIZED, 'Request from unknown host {}!'.format(str(req.remote_addr))))
                                raise falcon.HTTPError(falcon.HTTP_UNAUTHORIZED,
                                                       'Request from unknown host {}!'.format(str(req.remote_addr)))
                        else:
                            logger.debug('RESP {} {}'.format(falcon.HTTP_UNAUTHORIZED,
                                                             'Missing X-Auth-Token in the request header!'))
                            raise falcon.HTTPError(falcon.HTTP_UNAUTHORIZED, 'Missing X-Auth-Token in the request header!')
                    else:
                        #if get_keystone_token_credentials(keystone_token) is None:
                        #    logger.debug('RESP {} {}'.format(falcon.HTTP_UNAUTHORIZED, 'Invalid token provided!'))
                        #    raise falcon.HTTPError(falcon.HTTP_UNAUTHORIZED, 'Invalid token provided!')
                        headers['X-Auth-Token']=keystone_token

        body = ('data', req_body)

        #Important: Used by the inventory!
        if req.method in ('POST','PUT','PATCH') and 'application/json' in req.content_type:
            try:
                body = ('json', json.loads(req_body))
            except (ValueError, UnicodeDecodeError):
                #Let the consumer to take care of the request data!
                pass

        if str(req.method) in ('GET', 'PUT', 'POST', 'DELETE', 'PATCH'):
            data = requests.request(req.method.lower(), url, data=get_data(body, 'data'), json=get_data(body, 'json'),
                                    headers=headers, verify=False)
            resp.status = falcon.get_http_status(data.status_code)
            resp.content_type = 'application/json'
            resp.body = data.text.encode('utf-8', 'replace')
        elif str(req.method) == 'OPTIONS':
            resp.status = falcon.HTTP_200
            resp.content_type = 'application/text'
        else:
            raise falcon.HTTPError(falcon.HTTP_METHOD_NOT_ALLOWED,
                                   'Unknown request method {} for {}!'.format(str(req.method), str(req.path)))

        logger.debug('RESP {} {}\nHEADER: {}\nBODY: {}'.format(resp.status,resp.content_type,resp._headers,resp.body))
#end Sinker

def get_data(body, type):
    if body:
        return body[1] if body[0] == type else None

class MyWSGIRequestHandler(WSGIRequestHandler):
    def log_message(self, format, *args):
        logger.info(format % args)

def parse_json_file(config_file):
    if not os.path.isfile(config_file):
        logger.error('File ' + config_file + ' does not exists!')
        sys.exit(ERROR_CODE)
    try:
        with open(config_file) as data_file:
            return json.load(data_file)
    except IOError as e:
        logger.error('I/O error({}): {}'.format(e.errno, e.strerror))
    except:
        logger.error('Unexpected error: {}'.format(sys.exc_info()[0]))
        raise

def start_falcon(config_file):
    global config, logger, memcache_client

    config = parse_json_file(config_file)

    if 'log_level' in config:
        logger.setLevel(config['log_level'])

    if 'log_file' in config:
        fileHandler = logging.FileHandler(config['log_file'])
        fileHandler.setFormatter(logFormatter)
        logging.getLogger().addHandler(fileHandler)

    if not 'wrapper' in config:
        logger.error('The wrapper section is missing in config!')
        sys.exit(ERROR_CODE)

    if not 'authentication' in config:
        logger.error('The authentication method is missing in config!')
        sys.exit(ERROR_CODE)

    if config['authentication'] == 'keystone' and not 'keystone_url' in config:
        logger.error('Keystone URL is missing in config!')
        sys.exit(ERROR_CODE)

    memcache_client = memcache.Client(['localhost:11211'], debug=0)

    api = falcon.API()
    api.add_route('/authenticate', Authenticate())
    api.add_sink(SinkAdapter(), '/')

    return api

class StaticResource(object):
    def on_get(self, req, resp):
        resp.status = falcon.HTTP_200
        resp.content_type = 'text/html'
        with open('index.html', 'r') as f:
            resp.body = f.read()

def main():

    application = start_falcon('sslrl.json')

    #Add test route
    application.add_route('/', StaticResource())

    #Start the simple server for testing purposes
    my_ip = config['ip']
    my_port = config['port']
    server_keyfile = config['ssl_key']
    server_certfile = config['ssl_cert']
    logger.info('Starting server at https://{}:{}'.format(my_ip, my_port))
    httpd = simple_server.make_server(my_ip, my_port, application, handler_class=MyWSGIRequestHandler)
    httpd.socket = ssl.wrap_socket(httpd.socket, keyfile=server_keyfile, certfile=server_certfile, server_side=True)
    httpd.serve_forever()

if __name__ == '__main__':
   main()
